/*
 
 DESCRIPTION:
 GPIBCON is a simple utility that provides a terminal-based console for GPIB
 instruments communicating over a virtual serial port.  The display is managed
 with simple ANSI cursor control sequences.  An ASCII configuration file is used
 to define the commands available along with startup, shutdown, and read.  See
 the configuration file for details.
 
 By default the config file is opened in the same path as the program with the
 '.conf' extension.  An alternate config file can be specified to manage
 multiple instruments.
 
 The program is written in ANSI C and should compile without modification on most
 U/L*nux & mac OS platforms.  Provided the include file is in the same folder
 the command 'cc gpibcon.c -lpthread' will create the executable.
 
 GPIBCON communicates using a virtual serial port enumerated as a character
 device in /dev.
 
 This software is licensed under the MIT license:
 
 Copyright 2019 Brian Cornell.

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 of the Software, and to permit persons to whom the Software is furnished to do
 so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 SYNTAX:        run at command line without options.
 
 KNOWN BUGS:    1) Command prompt will appear to hang if a command is issued
                from the config file expecting a response when in fact the
                instrument does not send one.

 DEPENDENCIES:  gpibcon.h - program include file
 
 VERSION:  1.0, see build file for build #
 
 MODIFICATION HISTORY:
 
 13 Feb 19, BWC, first release.
 
 
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <limits.h>
#include <pthread.h>

#include "gpibcon.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Module data
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
FILE *GpibDev = NULL;                   //GPIB device file handle
gcmd *Index[SIZ_ARY_IDX];               //Index of first / last pointers
gcmd Dict[SIZ_ARY_DCT];                 //Command dictionary
char *prgName = STR_PRGNAME;
int gpibDspWait = FALSE,                //gpibDsp process control
    gpibDspAck = FALSE;                 //gpibDsp process control


/*Function main.  Controls program execution - initialization, run, shutdown.
 *  Spawns a thread to display console data (uses pthread).
 *
 * Returns:         return code; refer to 'RC_' definitions in include file.
 *
 * Parameters:      argc=argument count
 *                  argv=array of pointers to argument strings
 *
 */
int main(int argc, char *argv[])
{
    pthread_t id;
    int ReturnCode,
        counter,
        argc_c,
        cflags;

    while (TRUE)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Argument processing:  If no arguments display program info & exit.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_NOARG;
        if (argc < SIZ_ARGC_MIN)
        {
            printf(FMT_PRG_DESC, VERMAJOR, VERMINOR, VERBUILD);
            printf(FMT_PRG_SYNTX, prgName);
            for (counter = 0; (counter < SIZ_A_FLG); ++counter)\
                printf(FMT_PRG_CMDL, Flags[counter].str, Flags[counter].desc);

            break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Parse command line flags.  The outer loop walks thru the list of
         * arguments and the inner loop scans the array of Flags against each
         * argument.  When a flag is found its respective bit in cflags is set
         * and the value pointer in Flags is set to that parameter.
         *
         * Invalid flags are ignored.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (argc_c = 1, cflags = 0; (argc_c < argc); ++argc_c)
        {
            for (counter = 0; (counter < SIZ_A_FLG); ++counter)
                if (!(strncasecmp(Flags[counter].str, argv[argc_c], SIZ_CMD_FLAG)))
                {
                    Flags[counter].value = argv[argc_c];
                    break;
                }
            if (counter < SIZ_A_FLG) cflags |= (FLG_BIT_SET << counter);
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Load config file.  If the 'different' flag is set skip past the
         * flag portion of the value; otherwise set value to the program name.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_DIFF_CONF) Flags[FLG_A_DIFF_CONF].value += 2;
        else Flags[FLG_A_DIFF_CONF].value = argv[POS_ARG_PRG];
        if ((ReturnCode = loadCfg(Flags[FLG_A_DIFF_CONF].value, \
                                  (cflags & FLG_DIFF_CONF), Index, Dict))) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * If dump flag set, dump config & exit
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_DUMP_CONF)
        {
            dumpCfg(Index, Dict);
            ReturnCode = RC_SUCCESS;
            break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                        Open GPIB device
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_DEVOFAIL;
        if ((GpibDev = fopen(argv[argc - 1], FMODE_GPIB)) == NULL) break;

        //Install signal handlers
        ReturnCode = RC_SIGERROR;
        if (signal(SIGINT, sigHandler) == SIG_ERR) break;
        if (signal(SIGTERM, sigHandler) == SIG_ERR) break;
        if (signal(SIGHUP, sigHandler) == SIG_ERR) break;
        if (signal(SIGQUIT, sigHandler) == SIG_ERR) break;

        //Initialize GPIB interface
        ReturnCode = RC_INITFAIL;
        if (gpibInit(GpibDev, Index[DICT_INIT_F])) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Setup display and start display thread
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        scrSetup();
        ReturnCode = RC_THREAD;
        if (pthread_create(&id, NULL, (void *(*)(void *)) &gpibDsp, NULL)) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Enter console mmode; exit upon return.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = cmdProc();
        break;
    }

    //Shutdown
    if (GpibDev)
    {
        gpibOff(GpibDev, Index[DICT_OFF_F]);
        fclose(GpibDev);
    }

    //Exit messages
    if (ReturnCode > RC_NOARG) fprintf(stderr, FMT_ERR_STDERR, prgName,\
                                       retCode[ReturnCode], errno);
    printf(FMT_PRG_EXIT);

    return(ReturnCode);
} //End function main


/*Function cmdProc.  Console command processor.  Relies on access to module
 *  data.
 *
 * Returns:         non-zero on error
 *
 * Parameters:
 *
 */
int cmdProc(void)
{
    gcmd *req;                                  //User command request
    struct timespec ts;                         //Used for nanosleep
    char console[SIZ_BUF_CON],
        req_c[SIZ_BUF_CON],                     //Command portion of request
        req_v[SIZ_BUF_CON],                     //Value portion of request
        rsp[SIZ_BUF_RSP];                       //Command response buffer
    int int_c,                                  //Internal command array offset
        rc = RC_SUCCESS;                        //Return code

    ts.tv_sec = DLY_S_RD_C;                     //Delay between GPIB commands
    ts.tv_nsec = DLY_NS_RD_C;

    while (TRUE)
    {
        //Setup command line
        memset((void *) console, CHR_NULL, SIZ_BUF_CON);
        ansiCursorPos(DSP_POS_CMD_R, DSP_POS_CMD_C);
        printf(FMT_CMD_PR);
        ansiEraseLine();
        fflush(stdout);
        fgets(console, SIZ_BUF_CON, stdin);

        //Setup display for response
        ansiCursorPos(DSP_POS_RSP_R, DSP_POS_RSP_C);
        printf(FMT_RSP_PR);
        ansiEraseLine();
        fflush(stdout);

        //Parse command line
        strnl((unsigned char *) console, SIZ_BUF_CON);      //Strip CR/LF
        strlcpy(req_c, console, SIZ_BUF_CON);               //Copy for command
        strlrtrim((unsigned char *) req_c, SIZ_BUF_CON);    //Strip trailing text
        strlcpy(req_v, console, SIZ_BUF_CON);               //Copy for value
        strlnshift((unsigned char *) req_v, SIZ_BUF_CON);   //Strip leading text
        
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Process matching commands in dictionary
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (req = Index[DICT_CONF_F]; (req); req = req->next)
            if (!(strcasecmp(req_c, req->cfg))) break;

        if (req)                                            //Found
        {
            if ((req->usrv) && (!(strlen(req_v)))) printf(FMT_RSP_NOV);
            else
            {
                //Pause display thread so no GPIB device contention
                gpibDspWait = TRUE;
                while (!(gpibDspAck));                      //Wait until released

                fprintf(GpibDev, FMT_EXC_CMD, req->cmd, req_v);         //Write command
                if (req->crsp)                              //Only get response if expecting
                {
                    nanosleep(&ts, NULL);
                    fprintf(GpibDev, FMT_DSP_WR, Index[DICT_READ_F]->cmd);  //Get response
                    if (!(fgets(rsp, SIZ_BUF_RSP, GpibDev))) \
                        strlcpy(rsp, FMT_RSP_IOE, SIZ_BUF_RSP);
                }
                else rsp[0] = CHR_NULL;

                //Release display thread & print result
                gpibDspWait = FALSE;
                printf(FMT_EXC_RSP, rsp);
            }
        }
        else
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Process internal commands
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            for (int_c = 0; (int_c < SIZ_A_ICMD); ++int_c)
                if (!(strcasecmp(req_c, Icmd[int_c].str))) break;

            if (int_c == CMD_A_QUIT)
            {
                ansiCursorPos(DSP_POS_EXT_R, DSP_POS_EXT_C);
                break;
            }
            else if (int_c == CMD_A_CLS)
            {
                gpibDspWait = TRUE;
                while (!(gpibDspAck));                  //Wait until released
                scrSetup();
                gpibDspWait = FALSE;                    //Resume
            }
            else if (int_c == CMD_A_HELP)
            {
                gpibDspWait = TRUE;
                while (!(gpibDspAck));                  //Wait until released
                ansiClrScr();
                printf(FMT_HLP_DCM);

                //Display dictionary help
                for (req = Index[DICT_CONF_F]; (req); req = req->next)
                    printf(FMT_HLP_TXT, req->cfg, req->hlp);

                //Display internal help
                printf(FMT_HLP_ICM);
                for (int_c = 0; (int_c < SIZ_A_ICMD); ++int_c)
                    printf(FMT_HLP_TXT, Icmd[int_c].str, Icmd[int_c].desc);
                printf(FMT_HLP_END);                    //All done
                fgets(console, SIZ_BUF_CON, stdin);     //Wait for user
                scrSetup();                             //Redraw screen
                gpibDspWait = FALSE;                    //Resume
            }
            else printf(FMT_RSP_INV);
        }
    }

    return(rc);
} //End function cmdProc


/*Function loadCfg.  Parses config file, loads dictrionary, initializes pointers.
 *
 * Returns:         RC_SUCCESS if loaded, otherwise RC_BADCFG
 *
 * Parameters:      cfg=config file to load
 *                  cfg_d=FALSE to load default config; otherwise load cfg
 *                      without modification
 *                  idx=array of pointers to manage dictionary linked list
 *                  dct=config dictionary array
 *
 */
int loadCfg(char *cfg, int cfg_d, gcmd *idx[], gcmd *dct)
{
    FILE *GpibCfg;                          //Config file handle
    int NumObj,                             //Number of objects loaded
        line,                               //Config line number
        rc,                                 //Return code
        sz_init,
        sz_read,
        sz_cfg,
        sz_dsp,
        sz_off;
    char CfgFile[SIZ_BUF_FILE],             //Buffer for config file path
        *rec_init,
        *rec_read,
        *rec_cfg,
        *rec_dsp,
        *rec_off;

    for (GpibCfg = NULL, rc = RC_BADCFG; (TRUE);)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Open config file & initialize memory objects
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (strlcpy(CfgFile, cfg, SIZ_BUF_FILE) > SIZ_BUF_FILE) break;
        if (!(cfg_d))
            if (strlcat(CfgFile, STR_EXT_CFG, SIZ_BUF_FILE) > SIZ_BUF_FILE) break;
        if (!(GpibCfg = fopen(CfgFile, FMODE_CONF))) break;

        memset((void *) dct, CHR_NULL, (SIZ_ARY_DCT * sizeof(gcmd)));
        memset((void *) idx, CHR_NULL, (SIZ_ARY_IDX * sizeof(gcmd *)));
        rec_init = REC_NAM_INIT;
        rec_read = REC_NAM_READ;
        rec_cfg = REC_NAM_CFG;
        rec_dsp = REC_NAM_DSP;
        rec_off = REC_NAM_OFF;
        sz_init = strlen(rec_init);
        sz_read = strlen(rec_read);
        sz_cfg = strlen(rec_cfg);
        sz_dsp = strlen(rec_dsp);
        sz_off = strlen(rec_off);

        for (line = 0, NumObj = 0; ((fgets(CfgFile, SIZ_BUF_FILE, GpibCfg)) && \
                                    NumObj < SIZ_ARY_DCT); ++line)
        {
            //Get rid of trailing CR/LF
            strnl((unsigned char *) CfgFile, SIZ_BUF_FILE);
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Initialization record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (!(strncasecmp(CfgFile, rec_init, sz_init)))
            {
                //Setup pointers
                if (!(idx[DICT_INIT_F]))
                {
                    idx[DICT_INIT_F] = (dct + NumObj);
                    idx[DICT_INIT_L] = idx[DICT_INIT_F];
                }
                else
                {
                    (dct + NumObj)->prev = idx[DICT_INIT_L];
                    idx[DICT_INIT_L]->next = (dct + NumObj);
                    idx[DICT_INIT_L] = (dct + NumObj);
                }

                //Finish record setup
                idx[DICT_INIT_L]->type = REC_TYP_INIT;
                strlcpy(idx[DICT_INIT_L]->cmd, (CfgFile + sz_init), SIZ_BUF_CMD);
                ++NumObj;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Read record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_read, sz_read)))
            {
                //Setup pointers; only one record allowed so last found is used
                if (!(idx[DICT_READ_F]))
                {
                    idx[DICT_READ_F] = (dct + NumObj);
                    idx[DICT_READ_L] = idx[DICT_READ_F];
                    ++NumObj;
                }

                //Finish record setup
                idx[DICT_READ_F]->type = REC_TYP_READ;
                strlcpy(idx[DICT_READ_F]->cmd, (CfgFile + sz_read), SIZ_BUF_CMD);
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Config (command) record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_cfg, sz_cfg)))
            {
                //Setup pointers
                if (!(idx[DICT_CONF_F]))
                {
                    idx[DICT_CONF_F] = (dct + NumObj);
                    idx[DICT_CONF_L] = idx[DICT_CONF_F];
                }
                else
                {
                    (dct + NumObj)->prev = idx[DICT_CONF_L];
                    idx[DICT_CONF_L]->next = (dct + NumObj);
                    idx[DICT_CONF_L] = (dct + NumObj);
                }
                
                //Finish record setup
                idx[DICT_CONF_L]->type = REC_TYP_CFG;
                idx[DICT_CONF_L]->usrv = \
                strtol(getCfgVar(CfgFile, POS_CFG_PARM), NULL, BASE_TEN);
                idx[DICT_CONF_L]->crsp = \
                strtol(getCfgVar(CfgFile, POS_CFG_CRSP), NULL, BASE_TEN);
                strlcpy(idx[DICT_CONF_L]->cfg, \
                        getCfgVar(CfgFile, POS_CFG_USRC), SIZ_BUF_CFG);
                strlcpy(idx[DICT_CONF_L]->cmd, \
                        getCfgVar(CfgFile, POS_CFG_CMDC), SIZ_BUF_CMD);
                strlcpy(idx[DICT_CONF_L]->hlp, \
                        getCfgVar(CfgFile, POS_CFG_HELP), SIZ_BUF_HLP);
                ++NumObj;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Display record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_dsp, sz_dsp)))
            {
                //Setup pointers
                if (!(idx[DICT_DISP_F]))
                {
                    idx[DICT_DISP_F] = (dct + NumObj);
                    idx[DICT_DISP_L] = idx[DICT_DISP_F];
                }
                else
                {
                    (dct + NumObj)->prev = idx[DICT_DISP_L];
                    idx[DICT_DISP_L]->next = (dct + NumObj);
                    idx[DICT_DISP_L] = (dct + NumObj);
                }
                
                //Finish record setup
                idx[DICT_DISP_L]->type = REC_TYP_DSP;
                idx[DICT_DISP_L]->usrv = \
                strtol(getCfgVar(CfgFile, POS_DSP_ROW), NULL, BASE_TEN);
                idx[DICT_DISP_L]->crsp = \
                strtol(getCfgVar(CfgFile, POS_DSP_RLTX), NULL, BASE_TEN);
                strlcpy(idx[DICT_DISP_L]->cfg, \
                        getCfgVar(CfgFile, POS_DSP_LABL), SIZ_BUF_CFG);
                strlcpy(idx[DICT_DISP_L]->cmd, \
                        getCfgVar(CfgFile, POS_DSP_CMDC), SIZ_BUF_CMD);
                strlcpy(idx[DICT_DISP_L]->hlp, \
                        getCfgVar(CfgFile, POS_DSP_HELP), SIZ_BUF_HLP);
                ++NumObj;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Off (shutdown) record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_off, sz_off)))
            {
                //Setup pointers
                if (!(idx[DICT_OFF_F]))
                {
                    idx[DICT_OFF_F] = (dct + NumObj);
                    idx[DICT_OFF_L] = idx[DICT_OFF_F];
                }
                else
                {
                    (dct + NumObj)->prev = idx[DICT_OFF_L];
                    idx[DICT_OFF_L]->next = (dct + NumObj);
                    idx[DICT_OFF_L] = (dct + NumObj);
                }
                
                //Finish record setup
                idx[DICT_OFF_L]->type = REC_TYP_OFF;
                strlcpy(idx[DICT_OFF_L]->cmd, (CfgFile + sz_off), SIZ_BUF_CMD);
                ++NumObj;
            }
        }

        if (GpibCfg) fclose(GpibCfg);
        break;
    }

    //If all four are non-zero then valid minimum config
    if ((idx[DICT_INIT_F]) && (idx[DICT_READ_F]) && (idx[DICT_CONF_F]) && \
        (idx[DICT_DISP_F])) rc = RC_SUCCESS;

    return(rc);
} //End function loadCfg


/*Function dumpCfg.  Prints config file in logical order by section to stdout.
 *
 *
 * Returns:         n/a
 *
 * Parameters:      idx=array of linked list pointers
 *                  dct=config dictionary array
 *
 */
void dumpCfg(gcmd *idx[], gcmd *dct)
{
    gcmd *dsp;
    char *hdr = FMT_DMP_HDR;
    int nobj = 0;

    printf(hdr, REC_NAM_INIT);
    for (dsp = idx[DICT_INIT_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_INT, dsp->cmd);

    printf(hdr, REC_NAM_READ);
    for (dsp = idx[DICT_READ_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_RD, dsp->cmd);

    printf(hdr, REC_NAM_CFG);
    for (dsp = idx[DICT_CONF_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_CFG, dsp->cfg, dsp->cmd, dsp->usrv, dsp->crsp, dsp->hlp);

    printf(hdr, REC_NAM_DSP);
    for (dsp = idx[DICT_DISP_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_DSP, dsp->cfg, dsp->cmd, dsp->usrv, dsp->crsp, dsp->hlp);

    printf(hdr, REC_NAM_OFF);
    for (dsp = idx[DICT_OFF_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_OFF, dsp->cmd);

    printf(FMT_DMP_END, nobj, SIZ_ARY_DCT);
    return;
} //End function dumpCfg


/*Function gpibInit.  Initializes GPIB device.
 *
 *
 * Returns:         non-zero on error
 *
 * Parameters:  dev==pointer to object of tyep FILE for device handle
 *              rec=pointer to first initialization record in dictionary
 *
 */
int gpibInit(FILE *dev, gcmd *rec)
{
    struct timespec ts;
    
    ts.tv_sec = DLY_S_INT_C;
    ts.tv_nsec = DLY_NS_INT_C;
    while (rec)
    {
        fprintf(dev, FMT_INT_WR, rec->cmd);
        rec = rec->next;
        nanosleep(&ts, NULL);
    }
    
    return(RC_SUCCESS);
} //End function gpibInit


/*Function gpibOff.  Shutdown GPIB device.
 *
 * Returns:         non-zero on error
 *
 * Parameters:  dev==pointer to object of tyep FILE for device handle
 *              rec=pointer to first off record in dictionary
 *
 */
int gpibOff(FILE *dev, gcmd *rec)
{
    struct timespec ts;

    //Non-blocking attempt to prevent contention if display thread is running
    gpibDspWait = TRUE;
    ts.tv_sec = DLY_S_OFT_C;
    ts.tv_nsec = DLY_NS_OFT_C;
    nanosleep(&ts, NULL);

    ts.tv_sec = DLY_S_OFF_C;
    ts.tv_nsec = DLY_NS_OFF_C;
    while (rec)
    {
        fprintf(dev, FMT_OFF_WR, rec->cmd);
        rec = rec->next;
        nanosleep(&ts, NULL);
    }
    
    return(RC_SUCCESS);
} //End function gpibOff


/*Function strnl.  Removes the trailing New Line (NL) or Carriage Return (CR)
 *  from a string up to maxsz - 1 (whichever is encountered first) and replaces
 *  with a NULL byte.  In the event that buf was not NULL terminated and NL or
 *  CR aren't found buf will be left unterminated; therefore the calling
 *  function should check the return code.
 *
 *
 * Returns:     length of NULL terminated buffer (minus NULL byte) or zero if CR
 *              or LF isn't found.  If a NULL is encountered first the function
 *              will leave the buffer unmodified and return the length of NULL
 *              terminated buffer (minus NULL byte).
 *
 * Parameters:  buf=pointer to buffer
 *              maxsz=size of buffer
 *
 */
int strnl(unsigned char *buf, int maxsz)
{
    int pos;
    
    for (pos = 0; (pos < maxsz); ++pos)
    {
        if ((*(buf + pos) == CHR_CR) || (*(buf + pos) == CHR_LF) || \
            (*(buf + pos) == CHR_NULL)) break;
    }
    if (pos < maxsz)
    {
        *(buf + pos) = CHR_NULL;
        return(pos);
    }
    return(0);
} //End function strnl


/*Function strlnshift.  Removes the leading text & white space from a string and
 *  shifts the remaining text left.  The shift continues until NULL or until
 *  maxsz - 1.  In the event that no text was found to shift, buf will be
 *  truncated to a NULL string.  Useful to isolate a value from a command.
 *
 * Returns:     length of left shifted text minus NULL terminator
 *
 * Parameters:  buf=pointer to string
 *              maxsz=size of buffer
 *
 */
int strlnshift(unsigned char *buf, int maxsz)
{
    int pos,
        lpos;
    
    //In this loop lpos is flag for leading text to eliminate
    for (pos = 0, lpos = 0; ((pos < maxsz) && (*(buf + pos))); ++pos)
    {
        if (*(buf + pos) != CHR_SPACE)
        {
            //Found text to shift; else found leading text to delete
            if (lpos == 2) break; else lpos = 1;
        }
        //Found whitespace after leading text
        else if (lpos == 1) lpos = 2;
    }

    //Shift text left; pos has starting byte
    for (lpos = 0; ((pos < maxsz) && (*(buf + pos))); ++pos, ++lpos)
        *(buf + lpos) = *(buf + pos);
    //lpos will always be < maxsz
    *(buf + lpos) = CHR_NULL;
    return(lpos);
} //End function strlnshift


/*Function strlrtrim.  Trims leading whitespace and all whitespace & text after
 *  first non-whitespace encounter.  Useful to isolate the leading text in a
 *  command string.
 *
 * Returns:     length of trimmed string or zero if not trimmed
 *
 * Parameters:  buf=pointer to string
 *              maxsz=size of buffer
 *
 */
int strlrtrim(unsigned char *buf, int maxsz)
{
    int pos,
        txt_s,                                  //Starting position
        txt_e,                                  //Ending position
        txt_f;                                  //Found start flag
    char chr;                                   //Byte assignment
    
    //
    for (pos = 0, txt_s = 0, txt_e = 0, txt_f = FALSE; \
         ((pos < maxsz) && (*(buf + pos))); ++pos)
    {
        if ((*(buf + pos) != CHR_SPACE) && (!(txt_f)))
        {
            txt_s = pos;
            txt_f = TRUE;
        }
        if ((*(buf + pos) == CHR_SPACE) && (txt_f)) break;
    }
    if (txt_f) txt_e = (pos - 1);

    //Reset pos so if nothing to trim tell caller; otherwise with length.
    //Necessarily inefficent loop to avoid same source/dest address when there
    //isn't leading whitespace.
    pos = 0;
    if (txt_f)
    {
        for (pos = 0; (txt_s <= txt_e); ++txt_s, ++pos)
        {
            chr = *(buf + txt_s);
            *(buf + pos) = chr;
        }
        *(buf + pos) = CHR_NULL;                //NULL terminate
    }

    return(pos);
} //End function strlrtrim


/*Function sigHandler.  Signal handler for user interrupt & TERM signals.
 * Attempts to ensure that open files / devices are closed.
 *
 * Returns:     n/a
 *
 * Parameters:  signal=signal #; refer to signal.h.
 *
 */
void sigHandler(int signal)
{
    static int recurse = 1;
    
    if (recurse)
    {
        recurse = 0;
        if (GpibDev)
        {
            gpibOff(GpibDev, Index[DICT_OFF_F]);
            fclose(GpibDev);
        }

        fprintf(stderr, FMT_TRM_SIGNAL, prgName, signal);
        exit(RC_SIGEXIT);
    }
    
    return;
} //End function sigHandler


/*Function getCfgVar.  Parses a config record for the given variable.  Not
 * thread safe!
 *
 * Returns:     pointer to NULL terminated string; will be 0 length if variable
 *              not found, 0 length, or corrupt record
 *
 * Parameters:  rec=pointer to NULL terminated config record
 *              pos=offset-1 number representing variable position; if set to
 *              0 then the record identifier & first variable will be returned
 *
 */
const char *getCfgVar(char *rec, int pos)
{
    static char var[SIZ_BUF_VAR];
    int rec_off,
        var_off,
        pos_cur,
        fndc;

    memset((void *) var, CHR_NULL, SIZ_BUF_VAR);        //Clear
    for (rec_off = 0, var_off = 0, pos_cur = 0, fndc = 0; \
         (*(rec + rec_off)); ++rec_off)
    {
        if (pos_cur == pos)
        {
            if ((*(rec + rec_off) != CHR_NULL) && \
                (*(rec + rec_off) != CHR_SEMICOLON) && \
                (var_off < (SIZ_BUF_VAR - 1)))
            {
                var[var_off] = *(rec + rec_off);
                ++var_off;
                continue;
            }
            else break;
        }

        //First look for colon to find start of first variable
        if ((*(rec + rec_off) == CHR_COLON) && (!(fndc)))
        {
            fndc = TRUE;
            ++pos_cur;                                  //Found 1st variable
            continue;
        }

        if (*(rec + rec_off) == CHR_SEMICOLON) ++pos_cur;    //Found next variable
    }

    return((const char *) var);
} //End function getCfgVar


/*Function gpibDsp.  Reads values from GPIB device and displays using DISP
 * dictionary records.  Runs as a thread.
 *
 * Returns:     n/a
 *
 * Parameters:  args=pointer to argument passed by parent
 *
 */
void gpibDsp(void *args)
{
    gcmd *dsp;
    char *dsperr = STR_ERROR;
    int dsperr_sz,
        dsprow;
    struct timespec ts;

    dsperr_sz = strlen(dsperr);

    while (TRUE)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Simple process control.  If gpibDspWait is non-zero set gpibDspAck
         * non-zero and halt all I/O.  When gpibDspWait clears, clear gpibDspAck
         * and resume.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((gpibDspWait) && (!(gpibDspAck))) gpibDspAck = TRUE;
        else if (!(gpibDspWait) && ((gpibDspAck))) gpibDspAck = FALSE;
        else if (!(gpibDspWait) && (!(gpibDspAck)))
        {
            //Read GPIB data
            ts.tv_sec = DLY_S_RD_C;
            ts.tv_nsec = DLY_NS_RD_C;
            for (dsp = Index[DICT_DISP_F]; (dsp); dsp = dsp->next)
            {
                fprintf(GpibDev, FMT_DSP_WR, dsp->cmd);                 //Request cmd
                nanosleep(&ts, NULL);
                fprintf(GpibDev, FMT_DSP_WR, Index[DICT_READ_F]->cmd);  //Read cmd
                nanosleep(&ts, NULL);
                if ((fgets(dsp->rsp, SIZ_BUF_RSP, GpibDev)))
                {
                    if (dsp->crsp) strlnshift((unsigned char *) dsp->rsp, \
                                              SIZ_BUF_RSP);
                    strnl((unsigned char *) dsp->rsp, SIZ_BUF_RSP);
                }
                else strlcpy(dsp->rsp, dsperr, dsperr_sz);
            }

            //Display GPIB data
            fflush(stdout);
            flockfile(stdout);
            ansiCursorSav();
            ansiCursorPos(DSP_POS_VAL_R, DSP_POS_VAL_C);
            fflush(stdout);
            for (dsprow = 1, dsp = Index[DICT_DISP_F]; (dsp); dsp = dsp->next)
            {
                if (dsprow != dsp->usrv)
                {
                    dsprow = dsp->usrv;
                    ansiCursorPos((unsigned char) dsprow, DSP_POS_VAL_C);
                    ansiEraseLine();
                }
                fprintf(stdout, FMT_DSP_VAL, dsp->cfg, dsp->rsp);
            }
            ansiCursorRst();
            fflush(stdout);
            funlockfile(stdout);
        }

        //Sleep until next display update
        ts.tv_sec = DLY_S_DUP_C;
        ts.tv_nsec = DLY_NS_DUP_C;
        nanosleep(&ts, NULL);
    }

    return;
} //End function gpibDsp


/*Function ansiClrScr.  Clears terminal screen.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiClrScr(void)
{
    printf("%s%s", ANSIESC, ANSICLRSCR);
    return;
} //End function ansiClrScr


/*Function ansiEraseLine.  Erases line fromo cursor position to end of line.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiEraseLine(void)
{
    printf("%s%c", ANSIESC, ANSIELINE);
    return;
} //End function ansiEraseLine


/*Function ansiCursorSav.  Tells terminal to save terminal position.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiCursorSav(void)
{
    printf(ANSISAVE);
    return;
} //End function ansiCursorSav


/*Function ansiCursorRst.  Tells terminal to restore terminal position.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiCursorRst(void)
{
    printf(ANSIRESTORE);
    return;
} //End function ansiCursorRst


/*Function ansiCursorPos.  Sets terminal cursor position.
 *
 * Returns:     n/a
 *
 * Parameters:  row=terminal row
 *              col=terminal column
 *
 */
void ansiCursorPos(unsigned char row, unsigned char col)
{
    char srow[ACP_SZ_STR], scol[ACP_SZ_STR];
    
    if ((row < ACP_MAX_ROW) && (col < ACP_MAX_COL))
    {
        snprintf(srow, ACP_SZ_STR, "%hhu", row);
        snprintf(scol, ACP_SZ_STR, "%hhu", col);
        printf("%s%s;%s%c", ANSIESC, srow, scol, ANSICPOS);
    }
    return;
} //End function ansiCursorPos


/*Function scrSetup.  Sets up the console display.  If called with the display
 * thread active the caller is expected to pause display updates.
 *
 * Returns:     n/a
 *
 * Parameters:  n/a
 *
 */
void scrSetup(void)
{
    ansiClrScr();
    ansiCursorPos(24, 0);
    printf(FMT_PRG_DESC, VERMAJOR, VERMINOR, VERBUILD);
    fflush(stdout);

    return;
} //End function scrSetup



//End file.
